from flask import Blueprint, abort, redirect, render_template, request, url_for, current_app
import os

from CTFd.cache import clear_team_session, clear_user_session
from CTFd.exceptions import TeamTokenExpiredException, TeamTokenInvalidException
from CTFd.models import Brackets, TeamFieldEntries, TeamFields, Teams, db, Challenges
from CTFd.utils import config, get_config, validators
from CTFd.utils.crypto import verify_password
from CTFd.utils.decorators import authed_only, ratelimit, registered_only
from CTFd.utils.decorators.modes import require_team_mode
from CTFd.utils.decorators.visibility import (
    check_account_visibility,
    check_score_visibility,
)
from CTFd.utils.helpers import get_errors, get_infos
from CTFd.utils.humanize.words import pluralize
from CTFd.utils.user import get_current_user, get_current_user_attrs
from CTFd.utils.validators import ValidationError
from CTFd.config import Config
from CTFd.models import ChallengeFiles, Files, PageFiles, db
import qrcode
teams = Blueprint("teams", __name__)


@teams.route("/teams")
@check_account_visibility
@require_team_mode
def listing():
    q = request.args.get("q")
    field = request.args.get("field", "name")
    filters = []

    if field not in ("name", "affiliation", "website"):
        field = "name"

    if q:
        filters.append(getattr(Teams, field).like("%{}%".format(q)))

    teams = (
        Teams.query.filter_by(hidden=False, banned=False)
        .filter(*filters)
        .order_by(Teams.id.asc())
        .paginate(per_page=50)
    )

    args = dict(request.args)
    args.pop("page", 1)

    return render_template(
        "teams/teams.html",
        teams=teams,
        prev_page=url_for(request.endpoint, page=teams.prev_num, **args),
        next_page=url_for(request.endpoint, page=teams.next_num, **args),
        q=q,
        field=field,
    )


@teams.route("/teams/invite", methods=["GET", "POST"])
@registered_only
@require_team_mode
def invite():
    infos = get_infos()
    errors = get_errors()
    code = request.args.get("code")
    team_creation = get_config("team_creation")

    if code is None:
        abort(404)

    user = get_current_user_attrs()
    if user.team_id:
        errors.append("Vous êtes déjà dans une équipe. Vous ne pouvez pas en rejoindre une autre.")

    try:
        team = Teams.load_invite_code(code)
    except TeamTokenExpiredException:
        abort(403, description="Cette URL d'invitation a expiré")
    except TeamTokenInvalidException:
        abort(403, description="Cette URL d'invitation n'est pas valide")

    team_size_limit = get_config("team_size", default=0)

    if request.method == "GET":
        if team_size_limit:
            infos.append(
                "Les équipes sont limitées à {limit} membre{plural}".format(
                    limit=team_size_limit, plural=pluralize(number=team_size_limit)
                )
            )

        return render_template(
            "teams/invite.html", team=team, infos=infos, errors=errors, team_creation=team_creation
        )

    if request.method == "POST":
        if errors:
            return (
                render_template(
                    "teams/invite.html", team=team, infos=infos, errors=errors, team_creation=team_creation
                ),
                403,
            )

        if team_size_limit and len(team.members) >= team_size_limit:
            errors.append(
                "{name} a déjà atteint le nomre maximum de membres {limit}".format(
                    name=team.name, limit=team_size_limit
                )
            )
            return (
                render_template(
                    "teams/invite.html", team=team, infos=infos, errors=errors, team_creation=team_creation
                ),
                403,
            )

        user = get_current_user()
        user.team_id = team.id
        db.session.commit()

        clear_user_session(user_id=user.id)
        clear_team_session(team_id=team.id)

        return redirect(url_for("challenges.listing"))


@teams.route("/teams/join", methods=["GET", "POST"])
@authed_only
@require_team_mode
@ratelimit(method="POST", limit=10, interval=5)
def join():
    infos = get_infos()
    errors = get_errors()

    user = get_current_user_attrs()
    if user.team_id:
        errors.append("Vous êtes déjà dans une équipe. Vous ne pouvez pas en rejoindre une autre.")

    if request.method == "GET":
        team_size_limit = get_config("team_size", default=0)
        if team_size_limit:
            plural = "" if team_size_limit == 1 else "s"
            infos.append(
                "Les équipes sont limitées à {limit} membre{plural}".format(
                    limit=team_size_limit, plural=plural
                )
            )
       
        return render_template("teams/join_team.html", infos=infos, errors=errors,team_as_password= get_config('team_as_password'))

    if request.method == "POST":
        teamname = request.form.get("name")
        passphrase = request.form.get("password", "").strip()

        team = Teams.query.filter_by(name=teamname).first()

        if errors:
            return (
                render_template("teams/join_team.html", infos=infos, errors=errors),
                403,
            )

        if team and (verify_password(passphrase, team.password) or not get_config('team_as_password')):
            team_size_limit = get_config("team_size", default=1)
            if team_size_limit and len(team.members) >= team_size_limit:
                errors.append(
                    "{name} a déjà atteint la limite de membres {limit}".format(
                        name=team.name, limit=team_size_limit
                    )
                )
                return render_template(
                    "teams/join_team.html", infos=infos, errors=errors
                )

            user = get_current_user()
            user.team_id = team.id
            db.session.commit()

            if len(team.members) == 1:
                team.captain_id = user.id
                db.session.commit()

            clear_user_session(user_id=user.id)
            clear_team_session(team_id=team.id)

            return redirect(url_for("challenges.listing"))
        else:
            errors.append("Cette information est incorrecte")
            return render_template("teams/join_team.html", infos=infos, errors=errors)


@teams.route("/teams/new", methods=["GET", "POST"])
@authed_only
@require_team_mode
def new():
    infos = get_infos()
    errors = get_errors()

    if bool(get_config("team_creation", default=False)) is False:
        abort(
            403,
            description="La création d'équipe est désactivée. Veuillez rejoindre une équipe.",
        )

    num_teams_limit = int(get_config("num_teams", default=0))
    num_teams = Teams.query.filter_by(banned=False, hidden=False).count()
    if num_teams_limit and num_teams >= num_teams_limit:
        abort(
            403,
            description=f"Nombre maximum d'équipes atteint ({num_teams_limit}). Veuillez rejoindre une équipe existante.",
        )

    user = get_current_user_attrs()
    
    if user.team_id:
        errors.append("Vous êtes déjà dans une équipe. Vous ne pouvez pas en rejoindre une autre.")

    if request.method == "GET":
        team_size_limit = get_config("team_size", default=0)
        if team_size_limit:
            plural = "" if team_size_limit == 1 else "s"
            infos.append(
                "Les équipes sont limitées à {limit} membre{plural}".format(
                    limit=team_size_limit, plural=plural
                )
            )
        return render_template("teams/new_team.html", infos=infos, errors=errors)

    elif request.method == "POST":
        teamname = request.form.get("name", "").strip()
        passphrase = request.form.get("password", "").strip()
        teamcolor = request.form.get("color")

        website = request.form.get("website")
        affiliation = request.form.get("affiliation")
        country = request.form.get("country")
        bracket_id = request.form.get("bracket_id", None)

        user = get_current_user()

        existing_team = Teams.query.filter_by(name=teamname).first()
        if existing_team:
            errors.append("Ce nom d'équipe est déjà pris")
        if not teamname:
            errors.append("Ce nom d'équipe est invalide")

        # Process additional user fields
        fields = {}
        for field in TeamFields.query.all():
            fields[field.id] = field

        entries = {}
        for field_id, field in fields.items():
            value = request.form.get(f"fields[{field_id}]", "").strip()
            if field.required is True and (value is None or value == ""):
                errors.append("Veuillez fournir tous les champs requis")
                break

            if field.field_type == "boolean":
                entries[field_id] = bool(value)
            else:
                entries[field_id] = value

        if website:
            valid_website = validators.validate_url(website)
        else:
            valid_website = True

        if affiliation:
            valid_affiliation = len(affiliation) < 128
        else:
            valid_affiliation = True

        if bracket_id:
            valid_bracket = bool(
                Brackets.query.filter_by(id=bracket_id, type="teams").first()
            )
        else:
            if Brackets.query.filter_by(type="teams").count():
                valid_bracket = False
            else:
                valid_bracket = True

        if country:
            try:
                validators.validate_country_code(country)
                valid_country = True
            except ValidationError:
                valid_country = False
        else:
            valid_country = True

        if valid_website is False:
            errors.append("Les sites Web doivent avoir une URL appropriée commençant par http ou https")
        if valid_affiliation is False:
            errors.append("Veuillez fournir une affiliation plus courte")
        if valid_country is False:
            errors.append("Pays invalide")
        if valid_bracket is False:
            errors.append("Veuillez fournir une catégorie valide")

        if errors:
            return render_template("teams/new_team.html", errors=errors), 403

        # Hide the created team if admin are set hidden
        hidden = False
        if user.type == "admin" and get_config("admin_visible") is False:
            hidden = True

        team = Teams(
            name=teamname,
            password=passphrase,
            color=teamcolor,
            captain_id=user.id,
            hidden=hidden,
            bracket_id=bracket_id,
        )

        if website:
            team.website = website
        if affiliation:
            team.affiliation = affiliation
        if country:
            team.country = country

        db.session.add(team)
        db.session.commit()

        for field_id, value in entries.items():
            entry = TeamFieldEntries(field_id=field_id, value=value, team_id=team.id)
            db.session.add(entry)
        db.session.commit()

        user.team_id = team.id
        db.session.commit()
        
        data = request.url_root+"teams/invite?code="+team.get_invite_code()
        print("invite code:" + data)
      
        img = qrcode.make(data)

        upload_file = current_app.config.get("UPLOAD_FOLDER")
        print(upload_file)

        qr_path=str(team.id)+"TeamQrCode"+'_n.png'
        
        try:
            img.save(upload_file+"/"+qr_path)
        except:
            os.mkdir(upload_file)  
            img.save(upload_file+"/"+qr_path)
            
        model_args = {"type": "standard", "sha1sum": str(team.id)+"TeamQrCode", "location":qr_path}

        model = Files
        file_row = model(**model_args)
        db.session.add(file_row)
        db.session.commit()

        clear_user_session(user_id=user.id)
        clear_team_session(team_id=team.id)

        return redirect(url_for("challenges.listing"))


@teams.route("/team")
@authed_only
@require_team_mode
def private():
    infos = get_infos()
    errors = get_errors()

    user = get_current_user()
    if not user.team_id:
        return render_template("teams/team_enrollment.html")

    team_id = user.team_id

    team = Teams.query.filter_by(id=team_id).first_or_404()
    solves = team.get_solves()
    awards = team.get_awards()
    fails = team.get_fails()
    

    team.members.sort(key=get_member_score, reverse=True)
    potentialScore = {}
    for  member in team.members:
        potentialScore[member.id] = get_member_score(member)
    place = team.place
    score = team.get_score(admin=True)

    if config.is_scoreboard_frozen():
        infos.append("Le classement a été gelé! Vous pouvez encore soumettre des défis, le classement sera mis à jour à la fin de l'évènement.")
    
    return render_template(
        "teams/private.html",
        solves=solves,
        fails=fails,
        awards=awards,
        user=user,
        team=team,
        score=score,
        place=place,
        score_frozen=config.is_scoreboard_frozen(),
        infos=infos,
        errors=errors,
        potentialScore = potentialScore,
    )

def get_member_score(member):
    total_score = 0
    for solve in member.solves:
        challenge = Challenges.query.filter_by(id=solve.challenge_id).first_or_404()
        if(solve.value):
            total_score += int(solve.value)
        else:
            total_score += int(challenge.value)
    for fail in member.fails:
        challenge = Challenges.query.filter_by(id=fail.challenge_id).first_or_404()
        if(fail.value):
            total_score += int(fail.value)
        else:
            total_score += int(challenge.value)
  
    
    return total_score

@teams.route("/teams/<int:team_id>")
@check_account_visibility
@check_score_visibility
@require_team_mode
def public(team_id):
    infos = get_infos()
    errors = get_errors()
    team = Teams.query.filter_by(id=team_id, banned=False, hidden=False).first_or_404()
    solves = team.get_solves()
    awards = team.get_awards()
    
    place = team.place
    score = team.score
    
    if errors:
        return render_template("teams/public.html", team=team, errors=errors)

    if config.is_scoreboard_frozen():
        infos.append("Le classement a été gelé! Vous pouvez encore soumettre des défis, le classement sera mis à jour à la fin de l'évènement.")

    return render_template(
        "teams/public.html",
        solves=solves,
        awards=awards,
        team=team,
        score=score,
        place=place,
        score_frozen=config.is_scoreboard_frozen(),
        infos=infos,
        errors=errors,
    )
